@@ -15,6 +15,7 @@ from typing import (
     List,
     Mapping,
     MutableMapping,
+    Sequence,
     TypeVar,
     Union,
     cast,
@@ -23,6 +24,10 @@ from typing import (
 import xarray as xr
 from xarray import DataArray, Dataset
 from xarray.core.arithmetic import SupportsArithmetic
+from xarray.core.groupby import GroupBy
+from xarray.core.resample import Resample
+from xarray.core.rolling import Coarsen, Rolling
+from xarray.core.weighted import Weighted
 
 from .criteria import cf_role_criteria, coordinate_criteria, regex
 from .helpers import bounds_to_vertices
@@ -37,13 +42,7 @@ from .utils import (
 )
 
 #: Classes wrapped by cf_xarray.
-_WRAPPED_CLASSES = (
-    xr.core.resample.Resample,
-    xr.core.groupby.GroupBy,
-    xr.core.rolling.Rolling,
-    xr.core.rolling.Coarsen,
-    xr.core.weighted.Weighted,
-)
+_WRAPPED_CLASSES = (Resample, GroupBy, Rolling, Coarsen, Weighted)
 
 #:  `axis` names understood by cf_xarray
 _AXIS_NAMES = ("X", "Y", "Z", "T")
@@ -100,7 +99,7 @@ def apply_mapper(
 
         try:
             results = mapper(obj, key)
-        except KeyError as e:
+        except (KeyError, ValueError) as e:
             if error or "I expected only one." in repr(e):
                 raise e
             else:
@@ -255,10 +254,12 @@ def _get_axis_coord(obj: DataArray | Dataset, key: str) -> list[str]:
         )
 
     search_in = set()
-    if "coordinates" in obj.encoding:
-        search_in.update(obj.encoding["coordinates"].split(" "))
-    if "coordinates" in obj.attrs:
-        search_in.update(obj.attrs["coordinates"].split(" "))
+    attrs_or_encoding = ChainMap(obj.attrs, obj.encoding)
+    coordinates = attrs_or_encoding.get("coordinates", None)
+    # Handles case where the coordinates attribute is None
+    # This is used to tell xarray to not write a coordinates attribute
+    if coordinates:
+        search_in.update(coordinates.split(" "))
     if not search_in:
         search_in = set(obj.coords)
 
@@ -307,7 +308,12 @@ def _get_measure(obj: DataArray | Dataset, key: str) -> list[str]:
         attrs_or_encoding = ChainMap(da.attrs, da.encoding)
         if "cell_measures" in attrs_or_encoding:
             attr = attrs_or_encoding["cell_measures"]
-            measures = parse_cell_methods_attr(attr)
+            try:
+                measures = parse_cell_methods_attr(attr)
+            except ValueError as e:
+                raise ValueError(
+                    f"{var} has malformed cell_measures attribute {attr}."
+                ) from e
             if key in measures:
                 results.update([measures[key]])
 
@@ -455,21 +461,53 @@ _DEFAULT_KEY_MAPPERS: Mapping[str, tuple[Mapper, ...]] = {
 }
 
 
-def _guess_bounds_dim(da):
+def _guess_bounds_dim(da, dim=None, out_dim="bounds"):
     """
-    Guess bounds values given a 1D coordinate variable.
+    Guess bounds values given a 1D or 2D coordinate variable.
     Assumes equal spacing on either side of the coordinate label.
+    This is a coarse approximation, especially for 2D bounds on curvilinear grids.
     """
-    assert da.ndim == 1
+    if dim is None:
+        if da.ndim not in [1, 2]:
+            raise ValueError(
+                f"If dim is None, variable {da.name} must be 1D or 2D. Received {da.ndim}D variable instead."
+            )
+        dim = da.dims
+    if not isinstance(dim, str):
+        if len(dim) > 2:
+            raise NotImplementedError(
+                "Adding bounds with more than 2 dimensions is not supported."
+            )
+        elif len(dim) == 2:
+            daX = _guess_bounds_dim(da, dim[0]).rename(bounds="Xbnds")
+            daXY = _guess_bounds_dim(daX, dim[1]).rename(bounds="Ybnds")
+            return xr.concat(
+                [
+                    daXY.isel(Xbnds=0, Ybnds=0),
+                    daXY.isel(Xbnds=0, Ybnds=1),
+                    daXY.isel(Xbnds=1, Ybnds=1),
+                    daXY.isel(Xbnds=1, Ybnds=0),
+                ],
+                out_dim,
+            ).transpose(..., "bounds")
+        else:
+            dim = dim[0]
+    if dim not in da.dims:
+        (dim,) = da.cf.axes[dim]
+    if dim not in da.coords:
+        raise NotImplementedError(
+            "Adding bounds for unindexed dimensions is not supported currently."
+        )
 
-    dim = da.dims[0]
     diff = da.diff(dim)
     lower = da - diff / 2
     upper = da + diff / 2
-    bounds = xr.concat([lower, upper], dim="bounds")
+    bounds = xr.concat([lower, upper], dim=out_dim)
 
-    first = (bounds.isel({dim: 0}) - diff[0]).assign_coords({dim: da[dim][0]})
-    result = xr.concat([first, bounds], dim=dim)
+    first = (bounds.isel({dim: 0}) - diff.isel({dim: 0})).assign_coords(
+        {dim: da[dim][0]}
+    )
+    result = xr.concat([first, bounds], dim=dim).transpose(..., "bounds")
 
     return result
 
@@ -610,6 +648,7 @@ def _getitem(
     """
 
     obj = accessor._obj
+    all_bounds = obj.cf.bounds if isinstance(obj, Dataset) else {}
     kind = str(type(obj).__name__)
     scalar_key = isinstance(key, str)
 
@@ -627,7 +666,7 @@ def _getitem(
         if not isinstance(obj, DataArray) and scalar_key:
             bounds = set()
             for name in names:
-                bounds.update(obj.cf.bounds.get(name, []))
+                bounds.update(all_bounds.get(name, []))
             names = set(names) - bounds
         return names
 
@@ -1059,6 +1098,52 @@ class CFAccessor:
 
         return self._all_cell_measures
 
+    def curvefit(
+        self,
+        coords: str | DataArray | Iterable[str | DataArray],
+        func: Callable[..., Any],
+        reduce_dims: Hashable | Iterable[Hashable] = None,
+        skipna: bool = True,
+        p0: dict[str, Any] = None,
+        bounds: dict[str, Any] = None,
+        param_names: Sequence[str] = None,
+        kwargs: dict[str, Any] = None,
+    ):
+
+        if coords is not None:
+            if isinstance(coords, str):
+                coords = (coords,)
+            coords = [
+                apply_mapper(  # type: ignore
+                    [_single(_get_coords)], self._obj, v, error=False, default=[v]  # type: ignore
+                )[
+                    0
+                ]  # type: ignore
+                for v in coords
+            ]
+        if reduce_dims is not None:
+            if isinstance(reduce_dims, Hashable):
+                reduce_dims: Iterable[Hashable] = (reduce_dims,)  # type: ignore
+            reduce_dims = [
+                apply_mapper(  # type: ignore
+                    [_single(_get_dims)], self._obj, v, error=False, default=[v]  # type: ignore
+                )[
+                    0
+                ]  # type: ignore
+                for v in reduce_dims  # type: ignore
+            ]
+
+        return self._obj.curvefit(
+            coords=coords,
+            func=func,
+            reduce_dims=reduce_dims,
+            skipna=skipna,
+            p0=p0,
+            bounds=bounds,
+            param_names=param_names,
+            kwargs=kwargs,
+        )
+
     def _process_signature(
         self,
         func: Callable,
@@ -1239,7 +1324,7 @@ class CFAccessor:
         coords = self._obj.coords
         dims = self._obj.dims
 
-        def make_text_section(subtitle, attr, valid_values, default_keys=None):
+        def make_text_section(subtitle, attr, valid_values=None, default_keys=None):
 
             with warnings.catch_warnings():
                 warnings.simplefilter("ignore")
@@ -1260,11 +1345,12 @@ class CFAccessor:
             vardict = {key: vardict[key] for key in ordered_keys if key in vardict}
 
             # Keep only valid values (e.g., coords or data_vars)
-            vardict = {
-                key: set(value).intersection(valid_values)
-                for key, value in vardict.items()
-                if set(value).intersection(valid_values)
-            }
+            if valid_values is not None:
+                vardict = {
+                    key: set(value).intersection(valid_values)
+                    for key, value in vardict.items()
+                    if set(value).intersection(valid_values)
+                }
 
             # Star for keys with dims only, tab otherwise
             rows = [
@@ -1293,6 +1379,11 @@ class CFAccessor:
             text = f"CF Flag variable with mapping:\n\t{flag_dict!r}\n\n"
         else:
             text = ""
+
+        if self.cf_roles:
+            text += make_text_section("CF Roles", "cf_roles")
+            text += "\n"
+
         text += "Coordinates:"
         text += make_text_section("CF Axes", "axes", coords, _AXIS_NAMES)
         text += make_text_section("CF Coordinates", "coordinates", coords, _COORD_NAMES)
@@ -1337,6 +1428,7 @@ class CFAccessor:
         varnames = list(self.axes) + list(self.coordinates)
         varnames.extend(list(self.cell_measures))
         varnames.extend(list(self.standard_names))
+        varnames.extend(list(self.cf_roles))
 
         return set(varnames)
 
@@ -1412,14 +1504,18 @@ class CFAccessor:
                 ChainMap(da.attrs, da.encoding).get("cell_measures", "")
                 for da in obj.data_vars.values()
             ]
+        as_dataset = self._maybe_to_dataset().reset_coords()
 
         keys = {}
         for attr in set(all_attrs):
             try:
                 keys.update(parse_cell_methods_attr(attr))
             except ValueError:
+                bad_vars = list(
+                    as_dataset.filter_by_attrs(cell_measures=attr).data_vars.keys()
+                )
                 warnings.warn(
-                    f"Ignoring bad cell_measures attribute: {attr}.",
+                    f"Ignoring bad cell_measures attribute: {attr} on {bad_vars}.",
                     UserWarning,
                     stacklevel=2,
                 )
@@ -1461,6 +1557,40 @@ class CFAccessor:
 
         return {k: sorted(v) for k, v in vardict.items()}
 
+    @property
+    def cf_roles(self) -> dict[str, list[str]]:
+        """
+        Returns a dictionary mapping cf_role names to variable names.
+
+        Returns
+        -------
+        dict
+            Dictionary mapping cf_role names to variable names.
+
+        References
+        ----------
+        Please refer to the CF conventions document : http://cfconventions.org/Data/cf-conventions/cf-conventions-1.8/cf-conventions.html#coordinates-metadata
+
+        Examples
+        --------
+        >>> import cf_xarray
+        >>> from cf_xarray.datasets import dsg
+        >>> dsg.cf.cf_roles
+        {'profile_id': ['profile'], 'trajectory_id': ['trajectory']}
+        """
+        if isinstance(self._obj, Dataset):
+            variables = self._obj.variables
+        elif isinstance(self._obj, DataArray):
+            variables = self._obj.coords
+
+        vardict: dict[str, list[str]] = {}
+        for k, v in variables.items():
+            if "cf_role" in v.attrs:
+                role = v.attrs["cf_role"]
+                vardict[role] = vardict.setdefault(role, []) + [k]
+
+        return {k: sorted(v) for k, v in vardict.items()}
+
     def get_associated_variable_names(
         self, name: Hashable, skip_bounds: bool = False, error: bool = True
     ) -> dict[str, list[str]]:
@@ -1488,8 +1618,11 @@ class CFAccessor:
         coords: dict[str, list[str]] = {k: [] for k in keys}
         attrs_or_encoding = ChainMap(self._obj[name].attrs, self._obj[name].encoding)
 
-        if "coordinates" in attrs_or_encoding:
-            coords["coordinates"] = attrs_or_encoding["coordinates"].split(" ")
+        coordinates = attrs_or_encoding.get("coordinates", None)
+        # Handles case where the coordinates attribute is None
+        # This is used to tell xarray to not write a coordinates attribute
+        if coordinates:
+            coords["coordinates"] = coordinates.split(" ")
 
         if "cell_measures" in attrs_or_encoding:
             try:
@@ -1689,9 +1822,10 @@ class CFAccessor:
         DataArray or Dataset
             with appropriate attributes added
         """
-        obj = self._obj.copy(deep=True)
+        obj = self._obj.copy(deep=False)
         for var in obj.coords.variables:
-            if obj[var].ndim == 1 and _is_datetime_like(obj[var]):
+            var_is_coord = any(var in val for val in obj.cf.coordinates.values())
+            if not var_is_coord and obj[var].ndim == 1 and _is_datetime_like(obj[var]):
                 if verbose:
                     print(
                         f"I think {var!r} is of type 'time'. It has a datetime-like type."
@@ -1700,6 +1834,12 @@ class CFAccessor:
                 continue  # prevent second detection
 
             for name, pattern in regex.items():
+                var_is_axis = any(var in val for val in obj.cf.axes.values())
+                var_is_coord = any(var in val for val in obj.cf.coordinates.values())
+                if (name in _AXIS_NAMES and var_is_axis) or (
+                    name in _COORD_NAMES and var_is_coord
+                ):
+                    continue  # skip known axes/coordinates and prevent multiple guesses
                 # match variable names
                 if pattern.match(var.lower()):
                     if verbose:
@@ -1822,7 +1962,7 @@ class CFAccessor:
         info, table, aliases = parse_cf_standard_name_table(source)
 
         # Loop over standard names
-        ds = self._maybe_to_dataset().copy()
+        ds = self._maybe_to_dataset().copy(deep=False)
         attrs_to_print: dict = {}
         for std_name, var_names in ds.cf.standard_names.items():
 
@@ -2049,15 +2189,28 @@ class CFDatasetAccessor(CFAccessor):
         assert self._obj.sizes[bounds_dim] in [2, 4]
         return bounds_dim
 
-    def add_bounds(self, keys: str | Iterable[str]):
+    def add_bounds(
+        self,
+        keys: str | Iterable[str],
+        *,
+        dim: str | Iterable[str] | None = None,
+        output_dim: str = "bounds",
+    ):
         """
         Returns a new object with bounds variables. The bounds values are guessed assuming
-        equal spacing on either side of a coordinate label.
+        equal spacing on either side of a coordinate label. The linear estimation is only a
+        coarse approximation, especially 2D bounds on curvilinear grids. It is always better to use
+        bounds generated as part of the grid creation process. This method is purely for convenience.
 
         Parameters
         ----------
         keys : str or Iterable[str]
             Either a single variable name or a list of variable names.
+        dim : str or Iterable[str], optional
+            Core dimension(s) along which to estimate bounds. For 2D bounds, it can
+            be a list of 2 dimension names.
+        output_dim : str
+            The name of the bounds dimension to add.
 
         Returns
         -------
@@ -2091,7 +2244,7 @@ class CFDatasetAccessor(CFAccessor):
                 apply_mapper(_get_all, self._obj, key, error=False, default=[key])
             )
 
-        obj = self._maybe_to_dataset(self._obj.copy(deep=True))
+        obj = self._maybe_to_dataset(self._obj.copy(deep=False))
 
         bad_vars: set[str] = variables - set(obj.variables)
         if bad_vars:
@@ -2103,7 +2256,17 @@ class CFDatasetAccessor(CFAccessor):
             bname = f"{var}_bounds"
             if bname in obj.variables:
                 raise ValueError(f"Bounds variable name {bname!r} will conflict!")
-            obj.coords[bname] = _guess_bounds_dim(obj[var].reset_coords(drop=True))
+            out = _guess_bounds_dim(
+                obj[var].reset_coords(drop=True), dim=dim, out_dim=output_dim
+            )
+            if output_dim in obj.dims and (new := out[output_dim].size) != (
+                old := obj[output_dim].size
+            ):
+                raise ValueError(
+                    f"The `{output_dim}` dimension already exists but has a different length than the new one "
+                    f"({old} vs {new}). Please provide another bound dimension name with `output_dim`."
+                )
+            obj.coords[bname] = out
             obj[var].attrs["bounds"] = bname
 
         return self._maybe_to_dataarray(obj)
@@ -2166,7 +2329,7 @@ class CFDatasetAccessor(CFAccessor):
         else:
             coords = keys
 
-        obj = self._maybe_to_dataset(self._obj.copy(deep=True))
+        obj = self._maybe_to_dataset(self._obj.copy(deep=False))
 
         for coord in coords:
             try:
