@@ -58,6 +58,8 @@ coordinate_criteria: MutableMapping[str, MutableMapping[str, Tuple]] = {
             "Pressure",
         ),
         "axis": ("Z",),
+        "cartesian_axis": ("Z",),
+        "grads_dim": ("z",),
     },
     "vertical": {
         "standard_name": (
@@ -77,18 +79,27 @@ coordinate_criteria: MutableMapping[str, MutableMapping[str, Tuple]] = {
         "standard_name": ("projection_x_coordinate",),
         "_CoordinateAxisType": ("GeoX",),
         "axis": ("X",),
+        "cartesian_axis": ("X",),
+        "grads_dim": ("x",),
     },
     "Y": {
         "standard_name": ("projection_y_coordinate",),
         "_CoordinateAxisType": ("GeoY",),
         "axis": ("Y",),
+        "cartesian_axis": ("Y",),
+        "grads_dim": ("y",),
     },
-    "T": {"standard_name": ("time",), "_CoordinateAxisType": ("Time",), "axis": ("T",)},
-    "time": {
+    "T": {
         "standard_name": ("time",),
+        "_CoordinateAxisType": ("Time",),
+        "axis": ("T",),
+        "cartesian_axis": ("T",),
+        "grads_dim": ("t",),
     },
 }
 
+coordinate_criteria["time"] = coordinate_criteria["T"]
+
 # "long_name" and "standard_name" criteria are the same. For convenience.
 for coord, attrs in coordinate_criteria.items():
     coordinate_criteria[coord]["long_name"] = coordinate_criteria[coord][
