@@ -94,8 +94,10 @@ def _bounds_helper(values, n_core_dims, nbounds, order):
             top_left = values[..., -1:, :, 3]
             vertex_vals = np.block([[bot_left, bot_right], [top_left, top_right]])
         if order is None:  # We verify if the ccw version works.
-            calc_bnds = np.moveaxis(vertices_to_bounds(vertex_vals).values, 0, -1)
-            order = "counterclockwise" if np.all(calc_bnds == values) else "clockwise"
+            calc_bnds = vertices_to_bounds(vertex_vals).values
+            order = (
+                "counterclockwise" if np.allclose(calc_bnds, values) else "clockwise"
+            )
         if order == "clockwise":
             bot_left = values[..., :, :, 0]
             top_left = values[..., -1:, :, 1]
@@ -123,7 +125,7 @@ def vertices_to_bounds(
     Parameters
     ----------
     vertices : DataArray
-        The bounds to convert. Must be of shape (N, 2) or (N, M, 4).
+        The vertices to convert. Must be of shape (N + 1) or (N + 1, M + 1).
     out_dims : Sequence[str],
         The name of the dimension in the output. The first is the 'bounds'
         dimension and the following are the coordinate dimensions.
@@ -131,6 +133,7 @@ def vertices_to_bounds(
     Returns
     -------
     DataArray
+        Either of shape (2, N) or (4, N, M).
 
     References
     ----------
@@ -152,4 +155,6 @@ def vertices_to_bounds(
         raise ValueError(
             f"vertices format not understood. Got {vertices.dims} with shape {vertices.shape}."
         )
-    return xr.DataArray(bnd_vals, dims=out_dims[: vertices.ndim + 1])
+    return xr.DataArray(bnd_vals, dims=out_dims[: vertices.ndim + 1]).transpose(
+        ..., out_dims[0]
+    )
